#!/usr/bin/python2.5
# Copyright (C) 2008 One Laptop Per Child Association, Inc.
# Licensed under the terms of the GNU GPL v2 or later; see COPYING for details.
# Written by C. Scott Ananian <cscott@laptop.org>
"""Extensions to the sugar.bundle.activitybundle package, and resources for
discovering update sources."""
import os, os.path, re, urllib, urlparse
import bitfrost.util.urlrange as urlrange

import sugar.bundle.activitybundle as actbun
import sugar.env as env

USER_ACTIVITY_DIR=env.get_user_activities_path()
USER_GROUPS_FILE=os.path.join(USER_ACTIVITY_DIR, '.groups')
SYSTEM_ACTIVITY_DIR='/usr/share/sugar/activities'
SYSTEM_GROUPS_FILE='/etc/olpc-update/activity-groups'

DEFAULT_UPDATE_URL = "http://wiki.laptop.org/go/Activities"
"""For backwards-compatibility, hard-code a default update URLs for
already-distributed bundles that don't (yet) have them."""
DEFAULT_G1G1_GROUP = '%s/G1G1' % DEFAULT_UPDATE_URL
"""For backwards-compatibility with pre-703 builds, add a G1G1 group
by default."""

_DEBUG_VERSIONS=False
"""Hard-code sample version values for debugging."""

def _lookup_update_url(ab):
    # reparse configuration file.
    from ConfigParser import ConfigParser
    cp = ConfigParser()
    cp.readfp(ab.get_file('activity/activity.info'))
    section = 'Activity'
    # we've got a default update_url if none is specified
    ab._update_url = DEFAULT_UPDATE_URL
    # if an explicit update_url is listed, use that instead.
    if cp.has_option(section, 'update_url'):
        ab._update_url = cp.get(section, 'update_url')

def get_major_version():
    """Return the first two components of a dotted OLPC release number."""
    relnum = get_release_version()
    if relnum is None: return None # bail!
    major, nsubs = re.subn(r'^(\d+[.]\d+)[.]\d+$', r'\1', relnum, 1)
    return major if nsubs == 1 else None

def get_release_version():
    """Return the OLPC release number, which is a two- or three-part
    dotted version number like, "8.1"."""
    if _DEBUG_VERSIONS: return "8.1.1"
    try:
        return open('/etc/olpc-release').read().strip()
    except:
        return None
def get_os_version():
    """Return the "core OS" version, which is a integer optionally
    prefixed with a stream name, like "656" or "joyride 2100"."""
    if _DEBUG_VERSIONS: return "joyride 2100"
    try:
        return open('/boot/olpc_build').read().strip()
    except:
        return None

def url_variants(url):
    """Return a list of variants for a given URL, representing release
    version dependencies."""
    suffixes = [get_os_version(), get_release_version(), get_major_version()]
    return [urlparse.urljoin(url+'/', urllib.quote(s))
            for s in suffixes if s is not None] + [ url ]

def retrieve_first_variant(url, parse_func=(lambda f, u: f.read())):
    for u in url_variants(url):
        try:
            result = parse_func(urlrange.urlopen(u), u)
            if result is not None: return result
        except:
            continue # some failure, go on to next possible url
    return None # apparently, none of these URLs worked today.

# activity groups are represented by ~olpc/Activities/.groups
# which consists of a simple list of urls, one url per line.
def get_activity_group_urls():
    """Return a list of "group urls", which are the locations at which we
    can find group definitions.  Try a file in the home dir, then fall
    back to a build default, if present."""
    for filename in [ USER_GROUPS_FILE, SYSTEM_GROUPS_FILE ]:
        try:
            return [line.strip() for line in open(filename)
                    if not (line.strip()=='' or line.startswith('#'))]
        except IOError:
            continue # try next file.
    # hmm, no groups file.  let's make a special test to support g1g1
    # upgrades from 650/653/656.
    if 'org.laptop.WebActivity' not in \
           [act.get_bundle_id() for act in get_activities()]:
        return [ DEFAULT_G1G1_GROUP ]
    # otherwise, there really isn't a group.  Ho, hum.
    return []
    
def get_activities():
    """Return a list of all the activities on the system."""
    # XXX: we should use the ActivityRegistry, but it needs the gtk main
    #      loop, which we don't want to deal with at the moment.
    actlist = []
    for dname in (USER_ACTIVITY_DIR, SYSTEM_ACTIVITY_DIR):
        if not os.path.isdir(dname): continue # skip missing dirs
        for fname in os.listdir(dname):
            path = os.path.join(dname, fname)
            if not os.path.exists(os.path.join(path,'activity','activity.info')):
                continue # skip this, not really an activity.
            a = actbun.ActivityBundle(path)
            # XXX: HACK.  Add some useful methods to this object.
            def maker(ab, func):
                return lambda : func(ab)
            a.get_icon_data=maker(a, lambda self: open(self.get_icon()).read())
            a.get_update_url = maker(a, lambda self: self._update_url)
            _lookup_update_url(a)
            actlist.append(a)
    return actlist
