#!/usr/bin/python2.5
# Copyright (C) 2008 One Laptop Per Child Association, Inc.
# Licensed under the terms of the GNU GPL v2 or later; see COPYING for details.
# Written by C. Scott Ananian <cscott@laptop.org>
"""Tools for working with remote activity bundles."""

from ConfigParser import ConfigParser
from StringIO import StringIO
from zipfile import ZipFile

import bitfrost.util.urlrange as urlrange

def activity_info_from_url(url):
    """Return a `ConfigParser` for the activity.info file inside an .xo
    activity bundle at the location specified by `url`.  Attempts to use
    byte-range support on the target HTTP server to do this efficiently."""
    zf = ZipFile(urlrange.urlopen(url))
    return activity_info_from_zipfile(zf)

def activity_info_from_zipfile(zf):
    """Return a `ConfigParser` for the activity.info file inside the .xo
    activity bundle given by `zf`, which is a `ZipFile`."""
    cp = ConfigParser()
    filename = '%s/activity/activity.info' % activity_base_from_zipfile(zf)
    cp.readfp(StringIO(zf.read(filename)))
    return cp

def locale_activity_info_from_zipfile(zf):
    """Return a `ConfigParser` for the localized activity.linfo file inside
    the .xo activity bundle given by `zf`, which is a `ZipFile`."""
    import locale
    lang = locale.getdefaultlocale()[0]
    if lang is None: return None # bail early
    activity_base = activity_base_from_zipfile(zf)
    for f in ['locale/%s/activity.linfo' % lang,
              'locale/%s/activity.linfo' % lang[:2]]:
        filename = '%s/%s' % (activity_base, f)
        if filename not in zf.namelist(): continue
        cp = ConfigParser()
        cp.readfp(StringIO(zf.read(filename)))
        return cp
    return None # can't find a localized activity.info

def activity_base_from_zipfile(zf):
    """Return the activity directory base for the activity bundle given
    by the given `ZipFile` `zf`."""
    TRAILING_PART = '/activity/activity.info'
    n = [n for n in zf.namelist() if n.endswith(TRAILING_PART)]
    assert len(n)==1 # otherwise bogus
    return n[0][:-len(TRAILING_PART)]
    
def id_and_version_from_url(url):
    """Given the URL of a .xo file, return the activity id and
    integer version number, or (None,None) on error.

    Attempts to use byte-range support on the rarget HTTP server to extract
    just the activity/activity.info file and parse it.
    """
    try:
        cp = activity_info_from_url(url)
        SECTION = 'Activity'
        version = long(cp.get(SECTION, 'activity_version'))
        if cp.has_option(SECTION, 'bundle_id'):
            bundle_id = cp.get(SECTION, 'bundle_id')
        else:
            bundle_id = cp.get(SECTION, 'service_name')
        return (bundle_id, version)
    except:
        return None, None

